module TT_Find_Curves
  
  
  def self.run
    model = Sketchup.active_model
    
    start_time = Time.now
    Sketchup.status_text = 'Finding curves...'
    puts '=== Find Curves ==='
    
    curves = self.find_curves( model.active_entities )
    
    # Output performance data
    puts 'Method: TT_Find_Curves'
    puts "Found #{curves.size} curves in #{Time.now-start_time}s"
    
    # Visualize results
    materials = (0...50).map { |i|
      m=model.materials.add
      m.color = Sketchup::Color.new( 64+rand(128), 64+rand(128), 64+rand(128) )
      m
    }
    model.start_operation('Visualize Results', true)
    for curve in curves
      current_material = materials[ rand(10) ]
      for edge in curve
        edge.material = current_material
      end
    end
    model.commit_operation
    
    nil
  end # def
  
  
  # Find continous sets of edges not part of faces. The result is an array of
  # more arrays. Each sub-array contains a set of sorted edges.
  #
  # Sub-arrays may contains arrays of only one edge.
  #
  # @param [Array<Sketchup::Entity>] entities
  #
  # @return [Array<Array<Sketchup::Edge>>] An +Array+ of Arrays.
  # @since 2.5.0
  def self.find_curves( entities )
    t1=Time.now
    t_sort = 0
    
    t3 = Time.now
    # Use Hash for quick lookups
    cache = {}
    entities.each { |e|
      next unless e.is_a?( Sketchup::Edge ) && e.faces.empty?
      cache[e] = 0
    }
    filter = Time.now-t3
    
    curves = []
    until cache.empty?     
      curve = {} # Use Hash for quick lookups
      stack = [ cache.keys.first ]
      until stack.empty?
        # Fetch the next edges in the stack and add to the curve
        edge = stack.shift
        curve[edge] = 0
        cache.delete(edge)
        # Find next edges
        for v in edge.vertices
          vert_edges = v.edges
          next if vert_edges.size != 2
          for e in vert_edges
            stack << e unless curve.key?(e)
          end
        end # for v in
      end # until stack.empty?
      
      t2 = Time.now # Benchmark the sorting
      curves << self.sort( curve )
      t_sort += Time.now - t2
    end # until cache.empty?
    
    total=Time.now-t1
    sort_ratio = (t_sort/total.to_f*100).to_i
    puts "\n> Sorting: #{t_sort}s (#{sort_ratio}%)\n> Filter: #{filter}s\n\n"
    
    curves
  end
  
  
  # Sorts the given set of edges from start to end. If the edges form a loop
  # an arbitrary start is picked.
  #
  # @param [Hash] edges Sketchup::Edge as keys
  #
  # @return [Array<Sketchup::Edge>] Sorted set of edges.
  # @since 2.5.0
  def self.sort( edges ) #edges is a Hash
    # Get starting edge - then trace the connected edges from either end.
    start_edge = edges.keys.first
    
    # Find the next left and right edge
    vertices = start_edge.vertices
    
    left = []
    for e in vertices.first.edges
      left << e if e != start_edge && edges[e]
    end
    
    right = []
    for e in vertices.last.edges
      right << e if e != start_edge && edges[e]
    end
    
    return nil if left.size > 1 || right.size > 1 # Check for forks
    left = left.first
    right = right.first
    
    # Sort edges from start to end
    sorted = [start_edge]
    
    # Right
    edge = right
    until edge.nil?
      sorted << edge
      connected = []
      for v in edge.vertices
        for e in v.edges
          connected << e if edges[e] && !sorted.include?(e)
        end
      end
      return nil if connected.size > 1 # Check for forks
      edge = connected.first
    end
    
    # Left
    edge = left
    until edge.nil?
      sorted.unshift( edge )
      connected = []
      for v in edge.vertices
        for e in v.edges
          connected << e if edges[e] && !sorted.include?(e)
        end
      end
      return nil if connected.size > 1 # Check for forks
      edge = connected.first
    end
    
    sorted
  end

end # module